@@ -1,6 +1,6 @@
 /* dirmngr.c - Keyserver and X.509 LDAP access
  * Copyright (C) 2002 Klarälvdalens Datakonsult AB
- * Copyright (C) 2003-2004, 2006-2007, 2008, 2010-2011, 2020-2021 g10 Code GmbH
+ * Copyright (C) 2003, 2004, 2006, 2007, 2008, 2010, 2011 g10 Code GmbH
  * Copyright (C) 2014 Werner Koch
  *
  * This file is part of GnuPG.
@@ -17,7 +17,8 @@
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, see <https://www.gnu.org/licenses/>.
- * SPDX-License-Identifier: GPL-3.0-or-later
+ *
+ * SPDX-License-Identifier: GPL-3.0+
  */
 
 #include <config.h>
@@ -54,7 +55,6 @@
 #endif /*HTTP_USE_GNUTLS*/
 
 
-#define INCLUDED_BY_MAIN_MODULE 1
 #define GNUPG_COMMON_NEED_AFLOCAL
 #include "dirmngr.h"
 
@@ -99,7 +99,6 @@ enum cmd_and_opt_values {
   aFlush,
   aGPGConfList,
   aGPGConfTest,
-  aGPGConfVersions,
 
   oOptions,
   oDebug,
@@ -107,7 +106,6 @@ enum cmd_and_opt_values {
   oDebugWait,
   oDebugLevel,
   oGnutlsDebug,
-  oDebugCacheExpiredCerts,
   oNoGreeting,
   oNoOptions,
   oHomedir,
@@ -125,7 +123,6 @@ enum cmd_and_opt_values {
   oHTTPProxy,
   oLDAPProxy,
   oOnlyLDAPProxy,
-  oLDAPServer,
   oLDAPFile,
   oLDAPTimeout,
   oLDAPAddServers,
@@ -140,13 +137,10 @@ enum cmd_and_opt_values {
   oForce,
   oAllowOCSP,
   oAllowVersionCheck,
-  oStealSocket,
   oSocketName,
   oLDAPWrapperProgram,
   oHTTPWrapperProgram,
-  oIgnoreCert,
   oIgnoreCertExtension,
-  oIgnoreCRLExtension,
   oUseTor,
   oNoUseTor,
   oKeyServer,
@@ -158,7 +152,6 @@ enum cmd_and_opt_values {
   oConnectTimeout,
   oConnectQuickTimeout,
   oListenBacklog,
-  oCompatibilityFlags,
   aTest
 };
 
@@ -166,10 +159,6 @@ enum cmd_and_opt_values {
 
 static ARGPARSE_OPTS opts[] = {
 
-  ARGPARSE_c (aGPGConfList, "gpgconf-list", "@"),
-  ARGPARSE_c (aGPGConfTest, "gpgconf-test", "@"),
-  ARGPARSE_c (aGPGConfVersions, "gpgconf-versions", "@"),
-
   ARGPARSE_group (300, N_("@Commands:\n ")),
 
   ARGPARSE_c (aServer,   "server",  N_("run in server mode (foreground)") ),
@@ -178,99 +167,46 @@ static ARGPARSE_OPTS opts[] = {
   ARGPARSE_c (aSupervised,  "supervised", N_("run in supervised mode")),
 #endif
   ARGPARSE_c (aListCRLs, "list-crls", N_("list the contents of the CRL cache")),
-  ARGPARSE_c (aLoadCRL,  "load-crl", N_("|FILE|load CRL from FILE into cache")),
+  ARGPARSE_c (aLoadCRL,  "load-crl",  N_("|FILE|load CRL from FILE into cache")),
   ARGPARSE_c (aFetchCRL, "fetch-crl", N_("|URL|fetch a CRL from URL")),
   ARGPARSE_c (aShutdown, "shutdown",  N_("shutdown the dirmngr")),
   ARGPARSE_c (aFlush,    "flush",     N_("flush the cache")),
+  ARGPARSE_c (aGPGConfList, "gpgconf-list", "@"),
+  ARGPARSE_c (aGPGConfTest, "gpgconf-test", "@"),
 
-
-  ARGPARSE_header (NULL, N_("Options used for startup")),
-
-  ARGPARSE_s_n (oNoDetach, "no-detach", N_("do not detach from the console")),
-  ARGPARSE_s_n (oSh,       "sh",        N_("sh-style command output")),
-  ARGPARSE_s_n (oCsh,      "csh",       N_("csh-style command output")),
-  ARGPARSE_s_n (oStealSocket, "steal-socket", "@"),
-  ARGPARSE_s_s (oHomedir, "homedir", "@"),
-  ARGPARSE_conffile (oOptions,  "options", N_("|FILE|read options from FILE")),
-  ARGPARSE_noconffile (oNoOptions, "no-options", "@"),
-
-
-  ARGPARSE_header ("Monitor", N_("Options controlling the diagnostic output")),
+  ARGPARSE_group (301, N_("@\nOptions:\n ")),
 
   ARGPARSE_s_n (oVerbose,  "verbose",   N_("verbose")),
   ARGPARSE_s_n (oQuiet,    "quiet",     N_("be somewhat more quiet")),
-  ARGPARSE_s_n (oNoGreeting, "no-greeting", "@"),
+  ARGPARSE_s_n (oSh,       "sh",        N_("sh-style command output")),
+  ARGPARSE_s_n (oCsh,      "csh",       N_("csh-style command output")),
+  ARGPARSE_s_s (oOptions,  "options",   N_("|FILE|read options from FILE")),
   ARGPARSE_s_s (oDebugLevel, "debug-level",
                 N_("|LEVEL|set the debugging level to LEVEL")),
-  ARGPARSE_s_s (oDebug,    "debug", "@"),
-  ARGPARSE_s_n (oDebugAll, "debug-all", "@"),
-  ARGPARSE_s_i (oGnutlsDebug, "gnutls-debug", "@"),
-  ARGPARSE_s_i (oGnutlsDebug, "tls-debug", "@"),
-  ARGPARSE_s_i (oDebugWait, "debug-wait", "@"),
+  ARGPARSE_s_n (oNoDetach, "no-detach", N_("do not detach from the console")),
   ARGPARSE_s_s (oLogFile,  "log-file",
                 N_("|FILE|write server mode logs to FILE")),
-
-
-  ARGPARSE_header ("Configuration",
-                   N_("Options controlling the configuration")),
-
+  ARGPARSE_s_n (oBatch,    "batch",       N_("run without asking a user")),
+  ARGPARSE_s_n (oForce,    "force",       N_("force loading of outdated CRLs")),
+  ARGPARSE_s_n (oAllowOCSP, "allow-ocsp", N_("allow sending OCSP requests")),
   ARGPARSE_s_n (oAllowVersionCheck, "allow-version-check",
                 N_("allow online software version check")),
-  ARGPARSE_s_i (oListenBacklog, "listen-backlog", "@"),
-  ARGPARSE_s_i (oMaxReplies, "max-replies",
-                N_("|N|do not return more than N items in one query")),
-  ARGPARSE_s_u (oFakedSystemTime, "faked-system-time", "@"), /*(epoch time)*/
-  ARGPARSE_s_n (oDisableCheckOwnSocket, "disable-check-own-socket", "@"),
-  ARGPARSE_s_s (oIgnoreCert,"ignore-cert", "@"),
-  ARGPARSE_s_s (oIgnoreCertExtension,"ignore-cert-extension", "@"),
-  ARGPARSE_s_s (oIgnoreCRLExtension,"ignore-crl-extension", "@"),
-
-
-  ARGPARSE_header ("Network", N_("Network related options")),
-
-  ARGPARSE_s_n (oUseTor, "use-tor", N_("route all network traffic via Tor")),
-  ARGPARSE_s_n (oNoUseTor, "no-use-tor", "@"),
-  ARGPARSE_s_n (oDisableIPv4, "disable-ipv4", "@"),
-  ARGPARSE_s_n (oDisableIPv6, "disable-ipv6", "@"),
-  ARGPARSE_s_n (oStandardResolver, "standard-resolver", "@"),
-  ARGPARSE_s_n (oRecursiveResolver, "recursive-resolver", "@"),
-  ARGPARSE_s_i (oResolverTimeout, "resolver-timeout", "@"),
-  ARGPARSE_s_s (oNameServer, "nameserver", "@"),
-  ARGPARSE_s_i (oConnectTimeout, "connect-timeout", "@"),
-  ARGPARSE_s_i (oConnectQuickTimeout, "connect-quick-timeout", "@"),
-
-
-  ARGPARSE_header ("Keyserver", N_("Configuration for Keyservers")),
-
-  ARGPARSE_s_s (oKeyServer, "keyserver",
-                N_("|URL|use keyserver at URL")),
-  ARGPARSE_s_s (oHkpCaCert, "hkp-cacert",
-                N_("|FILE|use the CA certificates in FILE for HKP over TLS")),
-
-
-  ARGPARSE_header ("HTTP", N_("Configuration for HTTP servers")),
-
   ARGPARSE_s_n (oDisableHTTP, "disable-http", N_("inhibit the use of HTTP")),
+  ARGPARSE_s_n (oDisableLDAP, "disable-ldap", N_("inhibit the use of LDAP")),
   ARGPARSE_s_n (oIgnoreHTTPDP,"ignore-http-dp",
                 N_("ignore HTTP CRL distribution points")),
-  ARGPARSE_s_s (oHTTPProxy,  "http-proxy",
-                N_("|URL|redirect all HTTP requests to URL")),
-  ARGPARSE_s_n (oHonorHTTPProxy, "honor-http-proxy",
-                N_("use system's HTTP proxy setting")),
-  ARGPARSE_s_s (oLDAPWrapperProgram, "ldap-wrapper-program", "@"),
-
-
-  ARGPARSE_header ("LDAP", N_("Configuration of LDAP servers to use")),
-
-  ARGPARSE_s_n (oDisableLDAP, "disable-ldap", N_("inhibit the use of LDAP")),
   ARGPARSE_s_n (oIgnoreLDAPDP,"ignore-ldap-dp",
                 N_("ignore LDAP CRL distribution points")),
+  ARGPARSE_s_n (oIgnoreOCSPSvcUrl, "ignore-ocsp-service-url",
+                N_("ignore certificate contained OCSP service URLs")),
+
+  ARGPARSE_s_s (oHTTPProxy,  "http-proxy",
+                N_("|URL|redirect all HTTP requests to URL")),
   ARGPARSE_s_s (oLDAPProxy,  "ldap-proxy",
                 N_("|HOST|use HOST for LDAP queries")),
   ARGPARSE_s_n (oOnlyLDAPProxy, "only-ldap-proxy",
                 N_("do not use fallback hosts with --ldap-proxy")),
-  ARGPARSE_s_s (oLDAPServer, "ldapserver",
-                N_("|SPEC|use this keyserver to lookup keys")),
+
   ARGPARSE_s_s (oLDAPFile, "ldapserverlist-file",
                 N_("|FILE|read LDAP server list from FILE")),
   ARGPARSE_s_n (oLDAPAddServers, "add-servers",
@@ -279,12 +215,6 @@ static ARGPARSE_OPTS opts[] = {
   ARGPARSE_s_i (oLDAPTimeout, "ldaptimeout",
                 N_("|N|set LDAP timeout to N seconds")),
 
-
-  ARGPARSE_header ("OCSP", N_("Configuration for OCSP")),
-
-  ARGPARSE_s_n (oAllowOCSP, "allow-ocsp", N_("allow sending OCSP requests")),
-  ARGPARSE_s_n (oIgnoreOCSPSvcUrl, "ignore-ocsp-service-url",
-                N_("ignore certificate contained OCSP service URLs")),
   ARGPARSE_s_s (oOCSPResponder, "ocsp-responder",
                 N_("|URL|use OCSP responder at URL")),
   ARGPARSE_s_s (oOCSPSigner, "ocsp-signer",
@@ -293,21 +223,41 @@ static ARGPARSE_OPTS opts[] = {
   ARGPARSE_s_i (oOCSPMaxPeriod,    "ocsp-max-period", "@"),
   ARGPARSE_s_i (oOCSPCurrentPeriod, "ocsp-current-period", "@"),
 
+  ARGPARSE_s_i (oMaxReplies, "max-replies",
+                N_("|N|do not return more than N items in one query")),
 
-  ARGPARSE_header (NULL, N_("Other options")),
+  ARGPARSE_s_s (oNameServer, "nameserver", "@"),
+  ARGPARSE_s_s (oKeyServer, "keyserver", "@"),
+  ARGPARSE_s_s (oHkpCaCert, "hkp-cacert",
+                N_("|FILE|use the CA certificates in FILE for HKP over TLS")),
 
-  ARGPARSE_s_n (oForce,    "force",    N_("force loading of outdated CRLs")),
+  ARGPARSE_s_n (oUseTor, "use-tor", N_("route all network traffic via Tor")),
+  ARGPARSE_s_n (oNoUseTor, "no-use-tor", "@"),
+
+  ARGPARSE_s_n (oDisableIPv4, "disable-ipv4", "@"),
+  ARGPARSE_s_n (oDisableIPv6, "disable-ipv6", "@"),
 
   ARGPARSE_s_s (oSocketName, "socket-name", "@"),  /* Only for debugging.  */
-  ARGPARSE_s_n (oDebugCacheExpiredCerts, "debug-cache-expired-certs", "@"),
-  ARGPARSE_s_s (oCompatibilityFlags, "compatibility-flags", "@"),
 
-  ARGPARSE_header (NULL, ""),  /* Stop the header group.  */
-
-  /* Not yet used options.  */
-  ARGPARSE_s_n (oBatch,    "batch",       "@"),
+  ARGPARSE_s_u (oFakedSystemTime, "faked-system-time", "@"), /*(epoch time)*/
+  ARGPARSE_s_s (oDebug,    "debug", "@"),
+  ARGPARSE_s_n (oDebugAll, "debug-all", "@"),
+  ARGPARSE_s_i (oGnutlsDebug, "gnutls-debug", "@"),
+  ARGPARSE_s_i (oGnutlsDebug, "tls-debug", "@"),
+  ARGPARSE_s_i (oDebugWait, "debug-wait", "@"),
+  ARGPARSE_s_n (oDisableCheckOwnSocket, "disable-check-own-socket", "@"),
+  ARGPARSE_s_n (oNoGreeting, "no-greeting", "@"),
+  ARGPARSE_s_s (oHomedir, "homedir", "@"),
+  ARGPARSE_s_s (oLDAPWrapperProgram, "ldap-wrapper-program", "@"),
   ARGPARSE_s_s (oHTTPWrapperProgram, "http-wrapper-program", "@"),
-
+  ARGPARSE_s_n (oHonorHTTPProxy, "honor-http-proxy", "@"),
+  ARGPARSE_s_s (oIgnoreCertExtension,"ignore-cert-extension", "@"),
+  ARGPARSE_s_n (oStandardResolver, "standard-resolver", "@"),
+  ARGPARSE_s_n (oRecursiveResolver, "recursive-resolver", "@"),
+  ARGPARSE_s_i (oResolverTimeout, "resolver-timeout", "@"),
+  ARGPARSE_s_i (oConnectTimeout, "connect-timeout", "@"),
+  ARGPARSE_s_i (oConnectQuickTimeout, "connect-quick-timeout", "@"),
+  ARGPARSE_s_i (oListenBacklog, "listen-backlog", "@"),
 
   ARGPARSE_group (302,N_("@\n(See the \"info\" manual for a complete listing "
                          "of all commands and options)\n")),
@@ -332,13 +282,6 @@ static struct debug_flags_s debug_flags [] =
     { 77, NULL } /* 77 := Do not exit on "help" or "?".  */
   };
 
-/* The list of compatibility flags.  */
-static struct compatibility_flags_s compatibility_flags [] =
-  {
-    { 0, NULL }
-  };
-
-
 #define DEFAULT_MAX_REPLIES 10
 #define DEFAULT_LDAP_TIMEOUT 15  /* seconds */
 
@@ -378,10 +321,6 @@ static volatile int shutdown_pending;
 /* Flags to indicate that we shall not watch our own socket. */
 static int disable_check_own_socket;
 
-/* Flag indicating to start the daemon even if one already runs.  */
-static int steal_socket;
-
-
 /* Flag to control the Tor mode.  */
 static enum
   { TOR_MODE_AUTO = 0,  /* Switch to NO or YES         */
@@ -402,11 +341,6 @@ static int network_activity_seen;
 /* A list of filenames registred with --hkp-cacert.  */
 static strlist_t hkp_cacert_filenames;
 
-/* A flag used to clear the list of ldapservers iff --ldapserver is
- * given on the command line or one of the conf files. In this case we
- * want to clear all old specifications through the legacy
- * dirmngr_ldapservers.conf. */
-static int ldapserver_list_needs_reset;
 
 /* The timer tick used for housekeeping stuff.  The second constant is used when a shutdown is pending.  */
 #define TIMERTICK_INTERVAL           (60)
@@ -438,15 +372,11 @@ static npth_key_t my_tlskey_current_fd;
 /* Prototypes. */
 static void cleanup (void);
 #if USE_LDAP
-static ldap_server_t parse_ldapserver_file (const char* filename, int ienoent);
+static ldap_server_t parse_ldapserver_file (const char* filename);
 #endif /*USE_LDAP*/
-static fingerprint_list_t parse_fingerprint_item (const char *string,
-                                                  const  char *optionname,
-                                                  int want_binary);
+static fingerprint_list_t parse_ocsp_signer (const char *string);
 static void netactivity_action (void);
 static void handle_connections (assuan_fd_t listen_fd);
-static void gpgconf_versions (void);
-
 
 /* NPth wrapper function definitions. */
 ASSUAN_SYSTEM_NPTH_IMPL;
@@ -457,11 +387,9 @@ my_strusage( int level )
   const char *p;
   switch ( level )
     {
-    case  9: p = "GPL-3.0-or-later"; break;
     case 11: p = "@DIRMNGR@ (@GNUPG@)";
       break;
     case 13: p = VERSION; break;
-    case 14: p = GNUPG_DEF_COPYRIGHT_LINE; break;
     case 17: p = PRINTABLE_OS_NAME; break;
       /* TRANSLATORS: @EMAIL@ will get replaced by the actual bug
          reporting address.  This is so that we can change the
@@ -634,15 +562,6 @@ dirmngr_use_tor (void)
 }
 
 
-/* This is somewhat similar to dirmngr_use_tor but avoids a trial
- * connect and may thus be faster for this special case.  */
-int
-dirmngr_never_use_tor_p (void)
-{
-  return tor_mode == TOR_MODE_NEVER;
-}
-
-
 static void
 wrong_args (const char *text)
 {
@@ -699,14 +618,7 @@ parse_rereadable_options (ARGPARSE_ARGS *pargs, int reread)
           xfree (opt.ocsp_signer);
           opt.ocsp_signer = tmp;
         }
-      while (opt.ignored_certs)
-        {
-          fingerprint_list_t tmp = opt.ignored_certs->next;
-          xfree (opt.ignored_certs);
-          opt.ignored_certs = tmp;
-        }
       FREE_STRLIST (opt.ignored_cert_extensions);
-      FREE_STRLIST (opt.ignored_crl_extensions);
       http_register_tls_ca (NULL);
       FREE_STRLIST (hkp_cacert_filenames);
       FREE_STRLIST (opt.keyserver);
@@ -718,10 +630,6 @@ parse_rereadable_options (ARGPARSE_ARGS *pargs, int reread)
       set_dns_timeout (0);
       opt.connect_timeout = 0;
       opt.connect_quick_timeout = 0;
-      opt.ldaptimeout = DEFAULT_LDAP_TIMEOUT;
-      ldapserver_list_needs_reset = 1;
-      opt.debug_cache_expired_certs = 0;
-      opt.compat_flags = 0;
       return 1;
     }
 
@@ -773,8 +681,7 @@ parse_rereadable_options (ARGPARSE_ARGS *pargs, int reread)
     case oAllowVersionCheck: opt.allow_version_check = 1; break;
     case oOCSPResponder: opt.ocsp_responder = pargs->r.ret_str; break;
     case oOCSPSigner:
-      opt.ocsp_signer = parse_fingerprint_item (pargs->r.ret_str,
-                                                "--ocsp-signer", 0);
+      opt.ocsp_signer = parse_ocsp_signer (pargs->r.ret_str);
       break;
     case oOCSPMaxClockSkew: opt.ocsp_max_clock_skew = pargs->r.ret_int; break;
     case oOCSPMaxPeriod: opt.ocsp_max_period = pargs->r.ret_int; break;
@@ -796,32 +703,10 @@ parse_rereadable_options (ARGPARSE_ARGS *pargs, int reread)
       }
       break;
 
-    case oIgnoreCert:
-      {
-        fingerprint_list_t item, r;
-        item = parse_fingerprint_item (pargs->r.ret_str, "--ignore-cert", 20);
-        if (item)
-          {  /* Append  */
-            if (!opt.ignored_certs)
-              opt.ignored_certs = item;
-            else
-              {
-                for (r = opt.ignored_certs; r->next; r = r->next)
-                  ;
-                r->next = item;
-              }
-          }
-      }
-      break;
-
     case oIgnoreCertExtension:
       add_to_strlist (&opt.ignored_cert_extensions, pargs->r.ret_str);
       break;
 
-    case oIgnoreCRLExtension:
-      add_to_strlist (&opt.ignored_crl_extensions, pargs->r.ret_str);
-      break;
-
     case oUseTor:
       tor_mode = TOR_MODE_FORCE;
       break;
@@ -833,32 +718,6 @@ parse_rereadable_options (ARGPARSE_ARGS *pargs, int reread)
     case oStandardResolver: enable_standard_resolver (1); break;
     case oRecursiveResolver: enable_recursive_resolver (1); break;
 
-    case oLDAPServer:
-#if USE_LDAP
-      {
-        ldap_server_t server;
-        char *p;
-
-        p = pargs->r.ret_str;
-        if (!strncmp (p, "ldap:", 5) && !(p[5] == '/' && p[6] == '/'))
-          p += 5;
-
-        server = ldapserver_parse_one (p, NULL, 0);
-        if (server)
-          {
-            if (ldapserver_list_needs_reset)
-              {
-                ldapserver_list_needs_reset = 0;
-                ldapserver_list_free (opt.ldapservers);
-                opt.ldapservers = NULL;
-              }
-            server->next = opt.ldapservers;
-            opt.ldapservers = server;
-          }
-      }
-#endif
-      break;
-
     case oKeyServer:
       if (*pargs->r.ret_str)
         add_to_strlist (&opt.keyserver, pargs->r.ret_str);
@@ -880,23 +739,6 @@ parse_rereadable_options (ARGPARSE_ARGS *pargs, int reread)
       opt.connect_quick_timeout = pargs->r.ret_ulong * 1000;
       break;
 
-    case oLDAPTimeout:
-      opt.ldaptimeout = pargs->r.ret_int;
-      break;
-
-    case oDebugCacheExpiredCerts:
-      opt.debug_cache_expired_certs = 0;
-      break;
-
-    case oCompatibilityFlags:
-      if (parse_compatibility_flags (pargs->r.ret_str, &opt.compat_flags,
-                                     compatibility_flags))
-        {
-          pargs->r_opt = ARGPARSE_INVALID_ARG;
-          pargs->err = ARGPARSE_PRINT_WARNING;
-        }
-      break;
-
     default:
       return 0; /* Not handled. */
     }
@@ -913,7 +755,7 @@ parse_rereadable_options (ARGPARSE_ARGS *pargs, int reread)
 /* This fucntion is called after option parsing to adjust some values
  * and call option setup functions.  */
 static void
-post_option_parsing (enum cmd_and_opt_values cmd)
+post_option_parsing (void)
 {
   /* It would be too surpirsing if the quick timeout is larger than
    * the standard value.  */
@@ -921,18 +763,7 @@ post_option_parsing (enum cmd_and_opt_values cmd)
     opt.connect_quick_timeout = opt.connect_timeout;
 
   set_debug ();
-  /* For certain commands we do not want to set/test for Tor mode
-   * because that is somewhat expensive.  */
-  switch (cmd)
-    {
-    case aGPGConfList:
-    case aGPGConfTest:
-    case aGPGConfVersions:
-      break;
-    default:
-      set_tor_mode ();
-      break;
-    }
+  set_tor_mode ();
 }
 
 
@@ -971,7 +802,6 @@ static void
 thread_init (void)
 {
   npth_init ();
-  assuan_set_system_hooks (ASSUAN_SYSTEM_NPTH);
   gpgrt_set_syscall_clamp (npth_unprotect, npth_protect);
 
   /* Now with NPth running we can set the logging callback.  Our
@@ -992,10 +822,12 @@ main (int argc, char **argv)
   ARGPARSE_ARGS pargs;
   int orig_argc;
   char **orig_argv;
-  char *last_configname = NULL;
-  const char *configname = NULL;
+  FILE *configfp = NULL;
+  char *configname = NULL;
   const char *shell;
-  int debug_argparser = 0;
+  unsigned configlineno;
+  int parse_debug = 0;
+  int default_config =1;
   int greeting = 0;
   int nogreeting = 0;
   int nodetach = 0;
@@ -1045,6 +877,7 @@ main (int argc, char **argv)
   assuan_set_malloc_hooks (&malloc_hooks);
   assuan_set_assuan_log_prefix (log_get_prefix (NULL));
   assuan_set_gpg_err_source (GPG_ERR_SOURCE_DEFAULT);
+  assuan_set_system_hooks (ASSUAN_SYSTEM_NPTH);
   assuan_sock_init ();
   setup_libassuan_logging (&opt.debug, dirmngr_assuan_log_monitor);
 
@@ -1077,56 +910,63 @@ main (int argc, char **argv)
   orig_argv = argv;
   pargs.argc = &argc;
   pargs.argv = &argv;
-  pargs.flags= (ARGPARSE_FLAG_KEEP | ARGPARSE_FLAG_NOVERSION);
-  while (gnupg_argparse (NULL, &pargs, opts))
+  pargs.flags= 1|(1<<6);  /* do not remove the args, ignore version */
+  while (arg_parse( &pargs, opts))
     {
-      switch (pargs.r_opt)
+      if (pargs.r_opt == oDebug || pargs.r_opt == oDebugAll)
+        parse_debug++;
+      else if (pargs.r_opt == oOptions)
+        { /* Yes there is one, so we do not try the default one, but
+	     read the option file when it is encountered at the
+	     commandline */
+          default_config = 0;
+	}
+      else if (pargs.r_opt == oNoOptions)
+        default_config = 0; /* --no-options */
+      else if (pargs.r_opt == oHomedir)
         {
-        case oDebug:
-        case oDebugAll:
-          debug_argparser++;
-          break;
-        case oHomedir:
           gnupg_set_homedir (pargs.r.ret_str);
-          break;
         }
     }
-  /* Reset the flags.  */
-  pargs.flags &= ~(ARGPARSE_FLAG_KEEP | ARGPARSE_FLAG_NOVERSION);
 
   socket_name = dirmngr_socket_name ();
+  if (default_config)
+    configname = make_filename (gnupg_homedir (), DIRMNGR_NAME".conf", NULL );
 
-  /* The configuraton directories for use by gpgrt_argparser.  */
-  gnupg_set_confdir (GNUPG_CONFDIR_SYS, gnupg_sysconfdir ());
-  gnupg_set_confdir (GNUPG_CONFDIR_USER, gnupg_homedir ());
-
-  /* We are re-using the struct, thus the reset flag.  We OR the
-   * flags so that the internal intialized flag won't be cleared. */
   argc = orig_argc;
   argv = orig_argv;
   pargs.argc = &argc;
   pargs.argv = &argv;
-  pargs.flags |=  (ARGPARSE_FLAG_RESET
-                   | ARGPARSE_FLAG_KEEP
-                   | ARGPARSE_FLAG_SYS
-                   | ARGPARSE_FLAG_USER);
-  while (gnupg_argparser (&pargs, opts, DIRMNGR_NAME EXTSEP_S "conf"))
+  pargs.flags= 1;  /* do not remove the args */
+ next_pass:
+  if (configname)
     {
-      if (pargs.r_opt == ARGPARSE_CONFFILE)
+      configlineno = 0;
+      configfp = fopen (configname, "r");
+      if (!configfp)
         {
-          if (debug_argparser)
-            log_info (_("reading options from '%s'\n"),
-                      pargs.r_type? pargs.r.ret_str: "[cmdline]");
-          if (pargs.r_type)
+          if (default_config)
             {
-              xfree (last_configname);
-              last_configname = xstrdup (pargs.r.ret_str);
-              configname = last_configname;
-            }
+              if( parse_debug )
+                log_info (_("Note: no default option file '%s'\n"),
+                          configname );
+	    }
           else
-            configname = NULL;
-          continue;
-        }
+            {
+              log_error (_("option file '%s': %s\n"),
+                         configname, strerror(errno) );
+              exit(2);
+	    }
+          xfree (configname);
+          configname = NULL;
+	}
+      if (parse_debug && configname )
+        log_info (_("reading options from '%s'\n"), configname );
+      default_config = 0;
+    }
+
+  while (optfile_parse( configfp, configname, &configlineno, &pargs, opts) )
+    {
       if (parse_rereadable_options (&pargs, 0))
         continue; /* Already handled */
       switch (pargs.r_opt)
@@ -1141,7 +981,6 @@ main (int argc, char **argv)
         case aFetchCRL:
 	case aGPGConfList:
 	case aGPGConfTest:
-	case aGPGConfVersions:
           cmd = pargs.r_opt;
           break;
 
@@ -1151,11 +990,20 @@ main (int argc, char **argv)
 
         case oDebugWait: debug_wait = pargs.r.ret_int; break;
 
+        case oOptions:
+          /* Config files may not be nested (silently ignore them) */
+          if (!configfp)
+            {
+		xfree(configname);
+		configname = xstrdup(pargs.r.ret_str);
+		goto next_pass;
+	    }
+          break;
         case oNoGreeting: nogreeting = 1; break;
         case oNoVerbose: opt.verbose = 0; break;
+        case oNoOptions: break; /* no-options */
         case oHomedir: /* Ignore this option here. */; break;
         case oNoDetach: nodetach = 1; break;
-        case oStealSocket: steal_socket = 1; break;
         case oLogFile: logfile = pargs.r.ret_str; break;
         case oCsh: csh_style = 1; break;
         case oSh: csh_style = 0; break;
@@ -1165,6 +1013,9 @@ main (int argc, char **argv)
 #        endif /*USE_LDAP*/
           break;
 	case oLDAPAddServers: opt.add_new_ldapservers = 1; break;
+	case oLDAPTimeout:
+	  opt.ldaptimeout = pargs.r.ret_int;
+	  break;
 
         case oFakedSystemTime:
           gnupg_set_time ((time_t)pargs.r.ret_ulong, 0);
@@ -1178,26 +1029,20 @@ main (int argc, char **argv)
           listen_backlog = pargs.r.ret_int;
           break;
 
-        default:
-          if (configname)
-            pargs.err = ARGPARSE_PRINT_WARNING;
-          else
-            pargs.err = ARGPARSE_PRINT_ERROR;
-          break;
+        default : pargs.err = configfp? 1:2; break;
 	}
     }
-  gnupg_argparse (NULL, &pargs, NULL);  /* Release internal state.  */
-
-  if (!last_configname)
-    opt.config_filename = make_filename (gnupg_homedir (),
-                                         DIRMNGR_NAME EXTSEP_S "conf",
-                                         NULL);
-  else
+  if (configfp)
     {
-      opt.config_filename = last_configname;
-      last_configname = NULL;
+      fclose (configfp);
+      configfp = NULL;
+      /* Keep a copy of the name so that it can be read on SIGHUP. */
+      opt.config_filename = configname;
+      configname = NULL;
+      goto next_pass;
     }
-
+  xfree (configname);
+  configname = NULL;
   if (log_get_errorcount(0))
     exit(2);
   if (nogreeting )
@@ -1227,7 +1072,7 @@ main (int argc, char **argv)
           log_info (_("Note: '%s' is not considered an option\n"), argv[i]);
     }
 
-  if (!gnupg_access ("/etc/"DIRMNGR_NAME, F_OK)
+  if (!access ("/etc/"DIRMNGR_NAME, F_OK)
       && !strncmp (gnupg_homedir (), "/etc/", 5))
     log_info
       ("NOTE: DirMngr is now a proper part of %s.  The configuration and"
@@ -1245,22 +1090,20 @@ main (int argc, char **argv)
       log_printf ("\n");
     }
 
-  post_option_parsing (cmd);
+  post_option_parsing ();
 
-  /* Get LDAP server list from file unless --ldapserver has been used.  */
+  /* Get LDAP server list from file. */
 #if USE_LDAP
-  if (opt.ldapservers)
-    ;
-  else if (!ldapfile)
+  if (!ldapfile)
     {
       ldapfile = make_filename (gnupg_homedir (),
                                 "dirmngr_ldapservers.conf",
                                 NULL);
-      opt.ldapservers = parse_ldapserver_file (ldapfile, 1);
+      opt.ldapservers = parse_ldapserver_file (ldapfile);
       xfree (ldapfile);
     }
   else
-    opt.ldapservers = parse_ldapserver_file (ldapfile, 0);
+      opt.ldapservers = parse_ldapserver_file (ldapfile);
 #endif /*USE_LDAP*/
 
 #ifndef HAVE_W32_SYSTEM
@@ -1360,14 +1203,6 @@ main (int argc, char **argv)
           current_logfile = xstrdup (logfile);
         }
 
-      if (debug_wait)
-        {
-          log_debug ("waiting for debugger - my pid is %u .....\n",
-                     (unsigned int)getpid());
-          gnupg_sleep (debug_wait);
-          log_debug ("... okay\n");
-        }
-
 #ifndef HAVE_W32_SYSTEM
       if (strchr (socket_name, ':'))
         {
@@ -1417,11 +1252,7 @@ main (int argc, char **argv)
 #endif
               ))
 	{
-          /* Fixme: We should actually test whether a dirmngr is
-           * already running.  For now the steal option is a dummy. */
-          /* if (steal_socket) */
-          /*   log_info (N_("trying to steal socket from running %s\n"), */
-          /*             "dirmngr"); */
+          /* Fixme: We should test whether a dirmngr is already running. */
 	  gnupg_remove (redir_socket_name? redir_socket_name : socket_name);
 	  rc = assuan_sock_bind (fd, (struct sockaddr*) &serv_addr, len);
 	}
@@ -1432,7 +1263,7 @@ main (int argc, char **argv)
         {
           log_error (_("error binding socket to '%s': %s\n"),
                      serv_addr.sun_path,
-                     gpg_strerror (gpg_error_from_syserror ()));
+                     gpg_strerror (gpg_error_from_errno (errno)));
           assuan_sock_close (fd);
           dirmngr_exit (1);
         }
@@ -1631,25 +1462,78 @@ main (int argc, char **argv)
   else if (cmd == aGPGConfList)
     {
       unsigned long flags = 0;
+      char *filename;
       char *filename_esc;
 
+      /* First the configuration file.  This is not an option, but it
+	 is vital information for GPG Conf.  */
+      if (!opt.config_filename)
+        opt.config_filename = make_filename (gnupg_homedir (),
+                                             "dirmngr.conf", NULL );
+
+      filename = percent_escape (opt.config_filename, NULL);
+      es_printf ("gpgconf-dirmngr.conf:%lu:\"%s\n",
+              GC_OPT_FLAG_DEFAULT, filename);
+      xfree (filename);
+
+      es_printf ("verbose:%lu:\n", flags | GC_OPT_FLAG_NONE);
+      es_printf ("quiet:%lu:\n", flags | GC_OPT_FLAG_NONE);
       es_printf ("debug-level:%lu:\"none\n", flags | GC_OPT_FLAG_DEFAULT);
+      es_printf ("log-file:%lu:\n", flags | GC_OPT_FLAG_NONE);
+      es_printf ("force:%lu:\n", flags | GC_OPT_FLAG_NONE);
+
+      /* --csh and --sh are mutually exclusive, something we can not
+         express in GPG Conf.  --options is only usable from the
+         command line, really.  --debug-all interacts with --debug,
+         and having both of them is thus problematic.  --no-detach is
+         also only usable on the command line.  --batch is unused.  */
+
+      filename = make_filename (gnupg_homedir (),
+                                "dirmngr_ldapservers.conf",
+                                NULL);
+      filename_esc = percent_escape (filename, NULL);
+      es_printf ("ldapserverlist-file:%lu:\"%s\n", flags | GC_OPT_FLAG_DEFAULT,
+	      filename_esc);
+      xfree (filename_esc);
+      xfree (filename);
+
       es_printf ("ldaptimeout:%lu:%u\n",
-                 flags | GC_OPT_FLAG_DEFAULT, DEFAULT_LDAP_TIMEOUT);
+              flags | GC_OPT_FLAG_DEFAULT, DEFAULT_LDAP_TIMEOUT);
       es_printf ("max-replies:%lu:%u\n",
-                 flags | GC_OPT_FLAG_DEFAULT, DEFAULT_MAX_REPLIES);
+              flags | GC_OPT_FLAG_DEFAULT, DEFAULT_MAX_REPLIES);
+      es_printf ("allow-ocsp:%lu:\n", flags | GC_OPT_FLAG_NONE);
+      es_printf ("allow-version-check:%lu:\n", flags | GC_OPT_FLAG_NONE);
+      es_printf ("ocsp-responder:%lu:\n", flags | GC_OPT_FLAG_NONE);
+      es_printf ("ocsp-signer:%lu:\n", flags | GC_OPT_FLAG_NONE);
+
+      es_printf ("faked-system-time:%lu:\n", flags | GC_OPT_FLAG_NONE);
+      es_printf ("no-greeting:%lu:\n", flags | GC_OPT_FLAG_NONE);
+
+      es_printf ("disable-http:%lu:\n", flags | GC_OPT_FLAG_NONE);
+      es_printf ("disable-ldap:%lu:\n", flags | GC_OPT_FLAG_NONE);
+      es_printf ("honor-http-proxy:%lu\n", flags | GC_OPT_FLAG_NONE);
+      es_printf ("http-proxy:%lu:\n", flags | GC_OPT_FLAG_NONE);
+      es_printf ("ldap-proxy:%lu:\n", flags | GC_OPT_FLAG_NONE);
+      es_printf ("only-ldap-proxy:%lu:\n", flags | GC_OPT_FLAG_NONE);
+      es_printf ("ignore-ldap-dp:%lu:\n", flags | GC_OPT_FLAG_NONE);
+      es_printf ("ignore-http-dp:%lu:\n", flags | GC_OPT_FLAG_NONE);
+      es_printf ("ignore-ocsp-service-url:%lu:\n", flags | GC_OPT_FLAG_NONE);
+      /* Note: The next one is to fix a typo in gpgconf - should be
+         removed eventually. */
+      es_printf ("ignore-ocsp-servic-url:%lu:\n", flags | GC_OPT_FLAG_NONE);
+
+      es_printf ("use-tor:%lu:\n", flags | GC_OPT_FLAG_NONE);
 
       filename_esc = percent_escape (get_default_keyserver (0), NULL);
       es_printf ("keyserver:%lu:\"%s:\n", flags | GC_OPT_FLAG_DEFAULT,
                  filename_esc);
       xfree (filename_esc);
 
+
+      es_printf ("nameserver:%lu:\n", flags | GC_OPT_FLAG_NONE);
       es_printf ("resolver-timeout:%lu:%u\n",
                  flags | GC_OPT_FLAG_DEFAULT, 0);
     }
-  else if (cmd == aGPGConfVersions)
-    gpgconf_versions ();
-
   cleanup ();
   return !!rc;
 }
@@ -1706,8 +1590,6 @@ dirmngr_deinit_default_ctrl (ctrl_t ctrl)
 
   xfree (ctrl->http_proxy);
   ctrl->http_proxy = NULL;
-  nvc_release (ctrl->rootdse);
-  ctrl->rootdse = NULL;
 }
 
 
@@ -1727,7 +1609,7 @@ dirmngr_deinit_default_ctrl (ctrl_t ctrl)
 */
 #if USE_LDAP
 static ldap_server_t
-parse_ldapserver_file (const char* filename, int ignore_enoent)
+parse_ldapserver_file (const char* filename)
 {
   char buffer[1024];
   char *p;
@@ -1739,14 +1621,7 @@ parse_ldapserver_file (const char* filename, int ignore_enoent)
   fp = es_fopen (filename, "r");
   if (!fp)
     {
-      if (errno == ENOENT)
-        {
-          if (!ignore_enoent)
-            log_info ("No ldapserver file at: '%s'\n", filename);
-        }
-      else
-        log_error (_("error opening '%s': %s\n"), filename,
-                   strerror (errno));
+      log_error (_("error opening '%s': %s\n"), filename, strerror (errno));
       return NULL;
     }
 
@@ -1791,13 +1666,8 @@ parse_ldapserver_file (const char* filename, int ignore_enoent)
 }
 #endif /*USE_LDAP*/
 
-
-/* Parse a fingerprint entry as used by --ocsc-signer.  OPTIONNAME as
- * a description on the options used.  WANT_BINARY requests to store a
- * binary fingerprint.  Returns NULL on error and logs that error. */
 static fingerprint_list_t
-parse_fingerprint_item (const char *string,
-                        const char *optionname, int want_binary)
+parse_ocsp_signer (const char *string)
 {
   gpg_error_t err;
   char *fname;
@@ -1822,15 +1692,10 @@ parse_fingerprint_item (const char *string,
       if (j != 40 || !(spacep (string+i) || !string[i]))
         {
           log_error (_("%s:%u: invalid fingerprint detected\n"),
-                     optionname, 0);
+                     "--ocsp-signer", 0);
           xfree (item);
           return NULL;
         }
-      if (want_binary)
-        {
-          item->binlen = 20;
-          hex2bin (item->hexfpr, item->hexfpr, 20);
-        }
       return item;
     }
 
@@ -1913,12 +1778,6 @@ parse_fingerprint_item (const char *string,
           log_error (_("%s:%u: invalid fingerprint detected\n"), fname, lnr);
           errflag = 1;
         }
-      else if (want_binary)
-        {
-          item->binlen = 20;
-          hex2bin (item->hexfpr, item->hexfpr, 20);
-        }
-
       i++;
       while (spacep (p+i))
         i++;
@@ -1950,40 +1809,37 @@ static void
 reread_configuration (void)
 {
   ARGPARSE_ARGS pargs;
-  char *twopart;
+  FILE *fp;
+  unsigned int configlineno = 0;
   int dummy;
 
   if (!opt.config_filename)
     return; /* No config file. */
 
-  twopart = strconcat (DIRMNGR_NAME EXTSEP_S "conf" PATHSEP_S,
-                       opt.config_filename, NULL);
-  if (!twopart)
-    return;  /* Out of core.  */
+  fp = fopen (opt.config_filename, "r");
+  if (!fp)
+    {
+      log_error (_("option file '%s': %s\n"),
+                 opt.config_filename, strerror(errno) );
+      return;
+    }
 
   parse_rereadable_options (NULL, 1); /* Start from the default values. */
 
   memset (&pargs, 0, sizeof pargs);
   dummy = 0;
   pargs.argc = &dummy;
-  pargs.flags = (ARGPARSE_FLAG_KEEP
-                 |ARGPARSE_FLAG_SYS
-                 |ARGPARSE_FLAG_USER);
-  while (gnupg_argparser (&pargs, opts, twopart))
+  pargs.flags = 1;  /* do not remove the args */
+  while (optfile_parse (fp, opt.config_filename, &configlineno, &pargs, opts) )
     {
-      if (pargs.r_opt == ARGPARSE_CONFFILE)
-        {
-          log_info (_("reading options from '%s'\n"),
-                    pargs.r_type? pargs.r.ret_str: "[cmdline]");
-        }
-      else if (pargs.r_opt < -1)
-        pargs.err = ARGPARSE_PRINT_WARNING;
+      if (pargs.r_opt < -1)
+        pargs.err = 1; /* Print a warning. */
       else /* Try to parse this option - ignore unchangeable ones. */
         parse_rereadable_options (&pargs, 1);
     }
-  gnupg_argparse (NULL, &pargs, NULL);  /* Release internal state.  */
-  xfree (twopart);
-  post_option_parsing (0);
+  fclose (fp);
+
+  post_option_parsing ();
 }
 
 
@@ -1999,7 +1855,6 @@ dirmngr_sighup_action (void)
   crl_cache_deinit ();
   cert_cache_init (hkp_cacert_filenames);
   crl_cache_init ();
-  http_reinitialize ();
   reload_dns_stuff (0);
   ks_hkp_reload ();
 }
@@ -2089,7 +1944,6 @@ housekeeping_thread (void *arg)
   memset (&ctrlbuf, 0, sizeof ctrlbuf);
   dirmngr_init_default_ctrl (&ctrlbuf);
 
-  dns_stuff_housekeeping ();
   ks_hkp_housekeeping (curtime);
   if (network_activity_seen)
     {
@@ -2165,7 +2019,7 @@ handle_tick (void)
 
   /* Check whether the homedir is still available.  */
   if (!shutdown_pending
-      && gnupg_stat (gnupg_homedir (), &statbuf) && errno == ENOENT)
+      && stat (gnupg_homedir (), &statbuf) && errno == ENOENT)
     {
       shutdown_pending = 1;
       log_info ("homedir has been removed - shutting down\n");
@@ -2476,65 +2330,3 @@ dirmngr_get_current_socket_name (void)
   else
     return dirmngr_socket_name ();
 }
-
-
-
-/* Parse the revision part from the extended version blurb.  */
-static const char *
-get_revision_from_blurb (const char *blurb, int *r_len)
-{
-  const char *s = blurb? blurb : "";
-  int n;
-
-  for (; *s; s++)
-    if (*s == '\n' && s[1] == '(')
-      break;
-  if (*s)
-    {
-      s += 2;
-      for (n=0; s[n] && s[n] != ' '; n++)
-        ;
-    }
-  else
-    {
-      s = "?";
-      n = 1;
-    }
-  *r_len = n;
-  return s;
-}
-
-
-/* Print versions of dirmngr and used libraries.  This is used by
- * "gpgconf --show-versions" so that there is no need to link gpgconf
- * against all these libraries.  This is an internal API and should
- * not be relied upon.  */
-static void
-gpgconf_versions (void)
-{
-  const char *s;
-  int n;
-
-#if NPTH_VERSION_NUMBER >= 0x010800
-  es_fprintf (es_stdout, "* nPth %s (%s)\n\n",
-              npth_get_version (NULL), npth_get_version ("\x01\x02"));
-#endif /*NPTH_VERSION_NUMBER*/
-
-  s = get_revision_from_blurb (assuan_check_version ("\x01\x01"), &n);
-  es_fprintf (es_stdout, "* Libassuan %s (%.*s)\n\n",
-              assuan_check_version (NULL), n, s);
-
-  s = get_revision_from_blurb (ksba_check_version ("\x01\x01"), &n);
-  es_fprintf (es_stdout, "* KSBA %s (%.*s)\n\n",
-              ksba_check_version (NULL), n, s);
-
-#ifdef HTTP_USE_NTBTLS
-  s = get_revision_from_blurb (ntbtls_check_version ("\x01\x01"), &n);
-  es_fprintf (es_stdout, "* NTBTLS %s (%.*s)\n\n",
-              ntbtls_check_version (NULL), n, s);
-#elif HTTP_USE_GNUTLS
-  es_fprintf (es_stdout, "* GNUTLS %s\n\n",
-              gnutls_check_version (NULL));
-#endif
-
-}
